#ifndef _OBJECTS_H_
#define _OBJECTS_H_
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include "graphics.h"
#include "mouse.cpp"
#include "procs.h"
#include "defs.h"

#define RELEASED    0
#define PRESSED     1
#define SELECTED    2
#define OLDPRESSED  4

int MessageBox(char *message, int flags);
void InputBox(char *message, char *text);

struct TEvent {
  unsigned short command;
  char key;
  char scan;
  unsigned short x;
  unsigned short y;
  unsigned short b;
};

class TObject {
  protected:
    int x, y, w, h;
    int command;
    int flags, old_flags;
  public:
    TObject *next;
    TObject(int xx, int yy, int ww, int hh, int comm)
    {
      next = NULL;
      x = xx;
      y = yy;
      w = ww;
      h = hh;
      command = comm;
    }
    virtual void HandleEvent(TEvent& event);
    virtual void Draw() = 0;
};

class TBaseButton: public TObject {
  protected:
    int state;
  public:
    TBaseButton(int x, int y, int w, int h, int comm);
    virtual void HandleEvent(TEvent& event);
    virtual void Draw();
    ~TBaseButton();
};

void TObject::HandleEvent(TEvent& event)
{
  if (flags != old_flags)
  {
    mouse_hide();
    Draw();
    mouse_show();
  }
  old_flags = flags;
  event = event;
}

TBaseButton::TBaseButton(int x, int y, int w, int h, int comm):
             TObject(x, y, w, h, comm)
{
  state = RELEASED;
}

TBaseButton::~TBaseButton()
{
}

void TBaseButton::HandleEvent(TEvent& event)
{
  int old_state = state;

  if (event.x > x && event.x < x + w &&
      event.y > y && event.y < y + h)
  {
    state |= SELECTED;
    if (!(old_state & SELECTED) && (event.b == 1))
      state |= OLDPRESSED;
  }
  else
    state = RELEASED;

  if (state & SELECTED)
    if (event.b == 1)
    {
      if (!(state & OLDPRESSED))
        state |= PRESSED;
    }
    else
    {
      if (state & PRESSED)
      {
        state ^= PRESSED;
        event.command = command;
      }
      if (state & OLDPRESSED)
        state ^= OLDPRESSED;
    }

  if (old_state != state)
  {
    mouse_hide();
    Draw();
    mouse_show();
  }
}

void TBaseButton::Draw()
{
  if (state & SELECTED)
  {
    setcolor(BLACK);
    x++;
    y++;
    w -= 2;
    h -= 2;
    rectangle(x - 1, y - 1, x + w, y + h);
  }

  setfillstyle(1, GRAY);
  bar(x + 1, y + 1, x + w - 3, y + h - 3);

  if (state & PRESSED)
  {
    setcolor(GRAY);
    rectangle(x + 1, y + 1, x + w - 2, y + h - 2);
    setcolor(DARKGRAY);
    rectangle(x, y, x + w - 1, y + h - 1);
  }
  else
  {
    setcolor(WHITE);
    line(x, y, x + w - 2, y);
    line(x, y, x, y + h - 2);
    setcolor(BLACK);
    line(x + w - 1, y, x + w - 1, y + h - 1);
    line(x, y + h - 1, x + w - 1, y + h - 1);
    setcolor(DARKGRAY);
    line(x + 1, y + h - 2, x + w - 2, y + h - 2);
    line(x + w - 2, y + 1, x + w - 2, y + h - 2);
  }

  if (state & SELECTED)
  {
    setcolor(BLACK);
    x--;
    y--;
    w += 2;
    h += 2;
  }
}

class TBitBtn: public TBaseButton {
    int sprite;
    unsigned far *spr;
  public:
    TBitBtn(int x, int y, int w, int h, int comm, int sprite);
    TBitBtn(int x, int y, int w, int h, int comm, unsigned far *addr);
    virtual void HandleEvent(TEvent& event);
    virtual void Draw();
    ~TBitBtn();
};

TBitBtn::TBitBtn(int x, int y, int w, int h, int comm, int sprite):
         TBaseButton(x, y, w, h, comm)
{
  TBitBtn::sprite = sprite;
  spr = (unsigned *)(SPRITES_ADDR + sprite * SPRITE_SIZE);
  Draw();
}

TBitBtn::TBitBtn(int x, int y, int w, int h, int comm, unsigned far *addr):
         TBaseButton(x, y, w, h, comm)
{
  TBitBtn::sprite = sprite;
  spr = addr;
  Draw();
}

TBitBtn::~TBitBtn()
{
}

void TBitBtn::HandleEvent(TEvent& event)
{
  TBaseButton::HandleEvent(event);
}

void TBitBtn::Draw()
{
  int icon_x = x + w / 2 - (*spr + 1) / 2, icon_y = y + h / 2 - (*(spr + 1) + 1) / 2;

  if (state & PRESSED)
  {
    icon_x++;
    icon_y++;
  }

  TBaseButton::Draw();

  setcolor(DARKBLUE);
  putimage(icon_x, icon_y, spr, COPY_PUT);
}

class TTextButton: public TBaseButton {
    char *text;
    int text_height;
  public:
    TTextButton(int x, int y, int w, int h, int comm, char *caption);
    virtual void HandleEvent(TEvent& event);
    virtual void Draw();
    ~TTextButton();
};

TTextButton::TTextButton(int x, int y, int w, int h, int comm, char *caption):
             TBaseButton(x, y, w, h, comm)
{
  text = new char[strlen(caption) + 1];
  strcpy(text, caption);
  text_height = 8;
  Draw();
}

TTextButton::~TTextButton()
{
  delete text;
}

void TTextButton::HandleEvent(TEvent& event)
{
  TBaseButton::HandleEvent(event);
}

void TTextButton::Draw()
{
  int text_x = x + w / 2 - textwidth(text) / 2, text_y = y + h / 2 - 4;

  if (state & PRESSED)
  {
    text_x++;
    text_y++;
  }

  TBaseButton::Draw();

  setcolor(BLACK);
  settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
  outtextxy(text_x, text_y, text);
}

class TEditBox: public TObject {
    int state;
    int pressed;
    int x, y, b;
    int x1, y1, b1;
    void MouseDown(int x, int y);
    void MouseUp();
    void MouseMove(int x, int y);
    void FillLevel(int x, int y);
    void AddObject(int x, int y, int type);
    void DelObject(int x, int y);
    void ChangeProperty(int x, int y);
  public:
    int selected;
    TEditBox(int x, int y);
    virtual void HandleEvent(TEvent& event);
    virtual void Draw();
    void DrawObjects();
    void DrawSelected();
    ~TEditBox();
};

TEditBox::TEditBox(int x, int y):
          TObject(x, y, 320, 392, 0)
{
  pressed = False;
  b = b1 = 0;
  selected = 0;
  numobjects = 0;
}

void TEditBox::HandleEvent(TEvent& event)
{
  x = event.x;
  y = event.y;
  b = event.b;

  if ((x1 != x) || (y1 != y) || (b1 != b))
  if ((x > PANEL_X + 1) && (x < PANEL_X + 322) &&
      (x > PANEL_Y + 1) && (y < PANEL_Y + 394))
  {
    if (!b1 && b)
    {
      if ((event.command > 2000) && (event.command < 2018) && (event.command != 2014))
        if (event.command < 2014)
          AddObject(x, y, event.command - 2000);
        else
          AddObject(x, y, event.command - 1887);
      else if (event.command == 2014)
        DelObject(x, y);
      else if (event.command == CM_PROP)
        ChangeProperty(x, y);
      else
        MouseDown(x - PANEL_X - 2, y - PANEL_Y - 2);
    }
    if (b1 == b) MouseMove(x - PANEL_X - 2, y - PANEL_Y - 2);
  }
  if (b1 && !b) MouseUp();

  x1 = x;
  y1 = y;
  b1 = b;
}

void TEditBox::Draw()
{
  int i, j;

  for (i = 0; i < 16; i++)
    for (j = 0; j < 28; j++)
      if (Level[i][j] != 255)
        putimage(PANEL_X + 2 + i * 20, PANEL_Y + 2 + j * 14, SPRITES_ADDR + Level[i][j] * SPRITE_SIZE, COPY_PUT);
      else
      {
        setfillstyle(1, 0);
        bar(PANEL_X + 2 + i * 20, PANEL_Y + 2 + j * 14,
            PANEL_X + 2 + i * 20 + 19, PANEL_Y + 2 + j * 14 + 13);
      }
}

void TEditBox::FillLevel(int x, int y)
{
  Level[x / 20][y / 14] = selected;
  if (selected != 255)
    putimage(PANEL_X + 2 + x / 20 * 20,
             PANEL_Y + 2 + y / 14 * 14,
             SPRITES_ADDR + selected * SPRITE_SIZE,
             COPY_PUT);
  else
  {
    setfillstyle(1, 0);
    bar(PANEL_X + 2 + x / 20 * 20,
        PANEL_Y + 2 + y / 14 * 14,
        PANEL_X + 2 + x / 20 * 20 + 19,
        PANEL_Y + 2 + y / 14 * 14 + 13);
  }
  DrawObjects();
  modified = True;
}

void TEditBox::MouseDown(int x, int y)
{
  mouse_hide();
  pressed = True;
  FillLevel(x, y);
  mouse_show();
}

void TEditBox::MouseUp()
{
  pressed = False;
}

void TEditBox::MouseMove(int x, int y)
{
  if (!pressed) return;

  mouse_hide();
  FillLevel(x, y);
  mouse_show();
}

void TEditBox::DrawSelected()
{
  if (selected != 255)
    putimage(252, 453, SPRITES_ADDR + selected * SPRITE_SIZE, COPY_PUT);
  else
  {
    setfillstyle(1, 0);
    bar(252, 453, 271, 466);
  }
}

void TEditBox::AddObject(int x, int y, int type)
{
  if (numobjects >= MAX_OBJECTS) return;

  modified = True;

  Objects[numobjects].x = (x - TObject::x) / 20;
  Objects[numobjects].y = (y - TObject::y) / 14;
  Objects[numobjects].type = type;
  Objects[numobjects].reserved = 0;
  numobjects++;

  mouse_hide();
  DrawObjects();
  mouse_show();
}

void TEditBox::ChangeProperty(int x, int y)
{
  int i, j;
  char res[40];

  for (i = 0; i < numobjects; i++)
    if ((Objects[i].x == (x - TObject::x) / 20) &&
        (Objects[i].y == (y - TObject::y) / 14))
    {
      itoa(Objects[i].reserved, res, 10);
      InputBox("Change object property:", res);
      Objects[i].reserved = atoi(res);
      break;
    }

  modified = True;
}

void TEditBox::DelObject(int x, int y)
{
  int i, j;

  for (i = 0; i < numobjects; i++)
    if ((Objects[i].x == (x - TObject::x) / 20) &&
        (Objects[i].y == (y - TObject::y) / 14))
    {
      for (j = i; j < numobjects; j++)
        Objects[j] = Objects[j + 1];
      numobjects--;
      break;
    }

  modified = True;

  mouse_hide();
  Draw();
  DrawObjects();
  mouse_show();
}

void TEditBox::DrawObjects()
{
  int i;

  for (i = 0; i < numobjects; i++)
    if (Objects[i].type < 128)
      putimage(TObject::x + Objects[i].x * 20,
               TObject::y + Objects[i].y * 14,
               offss[Objects[i].type - 1], COPY_PUT);
    else
    {
      int u = TObject::x + Objects[i].x * 20;
      int v = TObject::y + Objects[i].y * 14;
      setcolor(WHITE);
      rectangle(u, v, u + 19, v + 13);
      switch (Objects[i].type)
      {
        case 128: line(u, v, u + 19, v + 13);
                  line(u + 19, v, u, v + 13);
                  break;
        case 129: moveto(u + 6, v + 5);
                  lineto(u + 10, v + 5);
                  lineto(u + 10, v + 2);
                  lineto(u + 14, v + 7);
                  lineto(u + 10, v + 12);
                  lineto(u + 10, v + 9);
                  lineto(u + 6, v + 9);
                  lineto(u + 6, v + 5);
                  break;
        case 130: moveto(u + 7, v + 3);
                  lineto(u + 7, v + 11);
                  lineto(u + 13, v + 11);
                  lineto(u + 13, v + 10);
                  break;
      }
    }
}

TEditBox::~TEditBox()
{
}

class TTextBox: public TObject {
    int maxlen;
  public:
    char *text;
    TTextBox(int x, int y, int w, int h, int command, int maxlen, char *text);
    ~TTextBox();
    virtual void HandleEvent(TEvent &event);
    virtual void Draw();
};

TTextBox::TTextBox(int x, int y, int w, int h, int command, int maxlen, char *text): TObject(x, y, w, h, command)
{
  TTextBox::maxlen = maxlen;
  TTextBox::text = new char[maxlen + 1];
  strcpy(TTextBox::text, text);
  Draw();
}

void TTextBox::Draw()
{
  int text_coord = h / 2 - 3;

  Draw3DHole(x, y, w, h, TEXTBOX);
  setcolor(BLACK);
  settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
  outtextxy(x + text_coord, y + text_coord, text);

  if ((flags & 2) == 2)
  {
    int n = x + text_coord + textwidth(text);
    line(n, y + 3, n, y + h - 4);
  }

  flags &= 251;
}

void TTextBox::HandleEvent(TEvent &event)
{
  char temp[2];

  if (event.x >= x && event.y >= y &&
      event.x < x + w && event.y < y + h)
  if (event.b == 1)
      flags |= 2;

  if (((flags & 2) == 2) && event.key && (event.key != 13))
  {
    if ((event.key == 8) && (strlen(text) != 0))
    {
      text[strlen(text) - 1] = 0;
      flags |= 4;
    }
    else if ((strlen(text) < maxlen) && (isalpha(event.key) || isdigit(event.key) ||
             (event.key == '.') || (event.key == '\\') || (event.key == ':')))
    {
      text[strlen(text) + 1] = 0;
      text[strlen(text)] = event.key;
      flags |= 4;
    }
  }

  TObject::HandleEvent(event);
  if (((flags & 2) == 2) && (event.key == 13)) event.command = command;
}

TTextBox::~TTextBox()
{
  delete text;
}

int MessageBox(char *message, int flags)
{
  TObject *b1 = NULL, *b2 = NULL, *b3 = NULL;
  int butt_flag = False;
  int xt, yt, bt;
  unsigned char k;
  TEvent event;

  mouse_hide();
  settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);

  int length = textwidth(message) + 20;
  if (length < 220) length = 220;

  char *buf = new char[imagesize((getmaxx() - length) / 2,
                                 getmaxy() / 2 - 30,
                                 (getmaxx() + length) / 2,
                                 getmaxy() / 2 + 30)];

  getimage((getmaxx() - length) / 2, getmaxy() / 2 - 30,
           (getmaxx() + length) / 2, getmaxy() / 2 + 30, buf);

  Draw3DBox((getmaxx() - length) / 2, getmaxy() / 2 - 30, length, 60);
  setcolor(BLACK);
  outtextxy((getmaxx() - textwidth(message)) / 2, getmaxy() / 2 - 20, message);

  switch (flags)
  {
    case MB_OK: b1 = new TTextButton(getmaxx() / 2 - 25, getmaxy() / 2, 50, 20, ID_OK, "OK");
                break;
    case MB_YESNO: b1 = new TTextButton(getmaxx() / 2 - 60, getmaxy() / 2, 50, 20, ID_YES, "Yes");
                   b2 = new TTextButton(getmaxx() / 2 + 10, getmaxy() / 2, 50, 20, ID_NO, "No");
                   break;
    case MB_YESNOCANCEL: b1 = new TTextButton(getmaxx() / 2 - 100, getmaxy() / 2, 60, 20, ID_YES, "Yes");
                         b2 = new TTextButton(getmaxx() / 2 - 30, getmaxy() / 2, 60, 20, ID_NO, "No");
                         b3 = new TTextButton(getmaxx() / 2 + 40, getmaxy() / 2, 60, 20, ID_CANCEL, "Cancel");
                         break;
  }

  mouse_show();

  do
  {
    event.command = 0;
    event.scan = 0;
    event.key = 0;

    asm
    {
      in     al,  60h
      mov    k,   al
    }

    if (kbhit())
    {
      event.key = getch();
      if (!event.key) event.scan = getch();
    }

    mouse_read(xt, yt, bt);
    event.x = xt;
    event.y = yt;
    event.b = bt;

    if (k == 57)
      butt_flag = True;
    if (k == 185)
      butt_flag = False;

    if (butt_flag) event.b = 1;

    if (b1) b1 -> HandleEvent(event);
    if (b2) b2 -> HandleEvent(event);
    if (b3) b3 -> HandleEvent(event);

    switch (event.scan)
    {
      case 77: mouse_hide(); mouse_set_pos(xt + 5, yt); mouse_show(); break;
      case 75: mouse_hide(); mouse_set_pos(xt - 5, yt); mouse_show(); break;
      case 72: mouse_hide(); mouse_set_pos(xt, yt - 5); mouse_show(); break;
      case 80: mouse_hide(); mouse_set_pos(xt, yt + 5); mouse_show(); break;
    }

    DrawMouse();
  } while (!event.command);

  mouse_hide();

  putimage((getmaxx() - length) / 2, getmaxy() / 2 - 30, buf, COPY_PUT);
  delete buf;
  if (b3) delete b3;
  if (b2) delete b2;
  if (b1) delete b1;

  mouse_show();

  return event.command;
}

void InputBox(char *message, char *text)
{
  TEvent event;
  int xt, yt, bt;
  int butt_flag = False;
  unsigned char k;

  mouse_hide();
  settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);

  int length = textwidth(message) + 20;
  if (length < 220) length = 220;

  char *buf = new char[imagesize((getmaxx() - length) / 2,
                                 getmaxy() / 2 - 30,
                                 (getmaxx() + length) / 2,
                                 getmaxy() / 2 + 30)];

  getimage((getmaxx() - length) / 2, getmaxy() / 2 - 30,
           (getmaxx() + length) / 2, getmaxy() / 2 + 30, buf);

  Draw3DBox((getmaxx() - length) / 2, getmaxy() / 2 - 30, length, 60);
  setcolor(BLACK);
  outtextxy((getmaxx() - textwidth(message)) / 2, getmaxy() / 2 - 20, message);

  TTextBox *t = new TTextBox(getmaxx() / 2 - 100, getmaxy() / 2, 200, 16, CM_EXIT, 20, text);

  mouse_show();

  do
  {
    event.command = 0;
    event.scan = 0;
    event.key = 0;

    asm
    {
      in     al,  60h
      mov    k,   al
    }

    mouse_read(xt, yt, bt);
    event.x = xt;
    event.y = yt;
    event.b = bt;

    if (kbhit())
    {
      event.key = getch();
      if (!event.key) event.scan = getch();
    }

    if (k == 57)
      butt_flag = True;
    if (k == 185)
      butt_flag = False;

    if (butt_flag) event.b = 1;

    t -> HandleEvent(event);

    switch (event.scan)
    {
      case 77: mouse_hide(); mouse_set_pos(xt + 5, yt); mouse_show(); break;
      case 75: mouse_hide(); mouse_set_pos(xt - 5, yt); mouse_show(); break;
      case 72: mouse_hide(); mouse_set_pos(xt, yt - 5); mouse_show(); break;
      case 80: mouse_hide(); mouse_set_pos(xt, yt + 5); mouse_show(); break;
    }

    DrawMouse();
  } while (!event.command);

  mouse_hide();

  putimage((getmaxx() - length) / 2, getmaxy() / 2 - 30, buf, COPY_PUT);
  strcpy(text, t -> text);
  delete buf;
  delete t;

  mouse_show();
}


#endif
